/*
 * Copyright 2013 Canonical Ltd.
 *
 * Authors:
 * Michael Frey: michael.frey@canonical.com
 * Matthew Fischer: matthew.fischer@canonical.com
 * Seth Forshee: seth.forshee@canonical.com
 *
 * This file is part of powerd.
 *
 * powerd is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * powerd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib.h>

#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <syslog.h>

#include <glib-object.h>
#include <gio/gio.h>
#include "powerd-internal.h"
#include "powerd-object.h"
#include "powerd-dbus.h"
#include "powerd-sensors.h"
#include "device-config.h"
#include "log.h"

#include <hybris/properties/properties.h>

#include "libsuspend.h"

#include <android/hardware/power.h>

static GThread *powerd_mainloop_thread;

namespace
{
static GMainLoop *main_loop = NULL;
static guint name_id;

static int g_exit_code = 0;

/* List with detected modems */
static GSList *g_modems = NULL;

static void free_call_data(void *data)
{
    struct call_data *call = (struct call_data *) data;
    g_free(call->obj_name);
    if (call->ofono_proxy)
        g_object_unref(call->ofono_proxy);
}

static gint call_data_cmp(gconstpointer a, gconstpointer b)
{
    const struct call_data *call = (const struct call_data *) a;
    const char *call_name = (const char *) b;

    return strcmp(call->obj_name, call_name);
}

/* List with existing calls */
static GSList *g_calls = NULL;

/* Flag set when proximity sensor is on */
static gboolean g_proximity_on = FALSE;

static struct power_module* _power_module;

static void
sigterm_quit(int signal)
{
    powerd_warn("SIGTERM recieved, cleaning up");
    powerd_exit(0);
}


}   //namespace

void powerd_hal_signal_activity(void)
{
    powerd_warn("signalling activity via HAL");
    if (_power_module && _power_module->powerHint)
        _power_module->powerHint(_power_module, POWER_HINT_INTERACTION, NULL);
}

/*
 * Once dbus is ready, force the screen on to make the system state
 * match the powerd iternal state
 */
void powerd_dbus_init_complete(void)
{
}

static void watch_modem(const char *modem)
{
    /* for incoming SMS signals */
    g_dbus_proxy_new_for_bus(G_BUS_TYPE_SYSTEM,
        G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
        NULL,
        "org.ofono",
        modem,
        "org.ofono.MessageManager",
        NULL,
        (GAsyncReadyCallback)ofono_proxy_connect_cb,
        NULL);

    /* for incoming calls Added/Removed signals */
    g_dbus_proxy_new_for_bus(G_BUS_TYPE_SYSTEM,
        G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
        NULL,
        "org.ofono",
        modem,
        "org.ofono.VoiceCallManager",
        NULL,
        (GAsyncReadyCallback)ofono_proxy_connect_cb,
        NULL);

    /* for USSD notifications/network requests */
    g_dbus_proxy_new_for_bus(G_BUS_TYPE_SYSTEM,
        G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
        NULL,
        "org.ofono",
        modem,
        "org.ofono.SupplementaryServices",
        NULL,
        (GAsyncReadyCallback)ofono_proxy_connect_cb,
        NULL);
}

void
ofono_get_modems_cb(GObject *source_object,
                    GAsyncResult *res,
                    gpointer user_data)
{
    GDBusProxy *client = G_DBUS_PROXY(source_object);
    GVariant *result;
    GVariant *item;
    GError *error = NULL;
    GVariantIter *iter;

    result = g_dbus_proxy_call_finish(client, res, &error);
    if (result == NULL) {
        powerd_warn("%s: call error", __func__);
        return;
    }

    g_variant_get(result, "(a(oa{sv}))", &iter);
    while ((item = g_variant_iter_next_value(iter))) {
        const char *obj_path;

        g_variant_get_child(item, 0, "&o", &obj_path);

        if (g_slist_find_custom(g_modems, obj_path, (GCompareFunc) strcmp)
                == NULL) {
            powerd_debug("active ofono modem %s", obj_path);
            g_modems = g_slist_prepend(g_modems, g_strdup(obj_path));
            watch_modem(obj_path);
        }

        g_variant_unref(item);
    }

    g_variant_unref(result);
}

static void notify_proximity_sensor(const char *call_state)
{
    /*
     * This function enables the proximity sensor in case the new state of the
     * call is:
     *
     * Active - meaning, the call is in progress --> Proximity On
     * Dialing - a signal made on outbound calls only which means that
     *           we have started to talk to the network, but the call has not
     *           been picked up. --> Proximity On (this signal is not
     *           sent on inbound calls)
     * Note that the proximity sensor is not activated for "incoming" state
     * For more details on states see voicecall-api.txt in the ofono source
     */

    if (g_proximity_on)
        return;

    if (strcmp("active", call_state) == 0
            || strcmp("dialing", call_state) == 0) {
        powerd_sensors_proximity_enable();
        g_proximity_on = TRUE;
    }
}

static void deactivate_proximity_sensor(void)
{
    if (!g_proximity_on)
        return;

    powerd_sensors_proximity_disable();
    g_proximity_on = FALSE;
}


void powerd_proximity_event(gboolean near)
{
    if (near)
        turn_display_on(FALSE, UNITY_SCREEN_REASON_PROXIMITY);
    else
        turn_display_on(TRUE, UNITY_SCREEN_REASON_PROXIMITY);
}

void
ofono_voicecall_get_props_cb(GObject *source_object, GAsyncResult *res,
                             gpointer user_data)
{
    GDBusProxy *client = G_DBUS_PROXY(source_object);
    GVariant *result;
    GVariant *item;
    GError *error = NULL;
    GVariantIter *iter = NULL;

    result = g_dbus_proxy_call_finish(client, res, &error);
    if (result == NULL) {
        powerd_warn("%s: call error", __func__);
        return;
    }

    g_variant_get(result, "(a{sv})", &iter);

    while ((item = g_variant_iter_next_value(iter))) {
        GVariant *value = NULL;
        const char *prop_name = NULL;
        const char *call_state = NULL;

        g_variant_get(item, "{&sv}", &prop_name, &value);

        if (strcmp("State", prop_name) == 0) {
            g_variant_get(value, "&s", &call_state);
            notify_proximity_sensor(call_state);
            powerd_info("GetProperties. Call State = %s", call_state);
        }

        g_variant_unref(value);
        g_variant_unref(item);
    }

    g_variant_unref(result);
}

void on_ofono_manager_signal(GDBusProxy *proxy, gchar *sender_name,
    gchar *signal_name, GVariant *parameters, gpointer user_data)
{
    const gchar *object_path;
    GVariant *tmp;
    char *modem;
    GSList *node;

    tmp = g_variant_get_child_value(parameters, 0);

    object_path = g_variant_get_string(tmp, NULL);
    node = g_slist_find_custom(g_modems, object_path, (GCompareFunc) strcmp);

    if (strcmp("ModemAdded", signal_name) == 0) {
        /* Add if not already in list */
        if (node == NULL) {
            powerd_debug("watching ofono modem %s", object_path);
            modem = g_strdup(object_path);
            g_modems = g_slist_prepend(g_modems, modem);
            watch_modem(modem);
        }
    } else if(strcmp("ModemRemoved", signal_name) == 0) {
        /* Remove if found in list */
        if (node != NULL) {
            powerd_debug("stop watch on ofono modem %s", object_path);
            g_free(g_modems->data);
            g_modems = g_slist_delete_link(g_modems, node);
        }
    }

    g_variant_unref(tmp);
}

void
on_ofono_voicecallmanager_signal (GDBusProxy *proxy,
           gchar      *sender_name,
           gchar      *signal_name,
           GVariant   *parameters,
           gpointer    user_data)
{
    GVariant *tmp;

    if (!strcmp(signal_name, "CallAdded")) {
        struct call_data *call;

        turn_display_on(TRUE, UNITY_SCREEN_REASON_NORMAL);

        call = (struct call_data *) calloc(1, sizeof(*call));
        tmp = g_variant_get_child_value(parameters, 0);
        call->obj_name = g_variant_dup_string(tmp, NULL);
        g_calls = g_slist_prepend(g_calls, call);

        powerd_info("%s incoming call", call->obj_name);

        /* Connect to voicecall interface */
        g_dbus_proxy_new_for_bus(G_BUS_TYPE_SYSTEM,
            G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
            NULL,
            "org.ofono",
            call->obj_name,
            "org.ofono.VoiceCall",
            NULL,
            (GAsyncReadyCallback) ofono_voicecall_proxy_connect_cb,
            call);

        g_variant_unref(tmp);

    } else if (!strcmp(signal_name, "CallRemoved")) {
        const gchar *object_path;
        GSList *node;

        tmp = g_variant_get_child_value(parameters, 0);
        object_path = g_variant_get_string(tmp, NULL);
        node = g_slist_find_custom(g_calls, object_path, call_data_cmp);
        if (node != NULL) {
            free_call_data(node->data);
            g_calls = g_slist_delete_link(g_calls, node);
            powerd_info("%s call removed", object_path);
        }

        if (g_calls == NULL)
            deactivate_proximity_sensor();

        g_variant_unref(tmp);
    }
}

void
on_ofono_voicecall_signal (GDBusProxy *proxy,
           gchar      *sender_name,
           gchar      *signal_name,
           GVariant   *parameters,
           gpointer    user_data)
{
    GVariant *value = NULL;
    const char *prop_name = NULL;
    const char *call_state = NULL;

    if (!strcmp(signal_name, "PropertyChanged")) {
        g_variant_get(parameters, "(&sv)", &prop_name, &value);

        if (!strcmp(prop_name, "State")) {
            g_variant_get(value, "&s", &call_state);

            powerd_info("Call State = %s", call_state);
            notify_proximity_sensor(call_state);
        }

        g_variant_unref(value);
    }
}

void
on_ofono_messagemanager_signal (GDBusProxy *proxy,
           gchar      *sender_name,
           gchar      *signal_name,
           GVariant   *parameters,
           gpointer    user_data)
{
    powerd_debug("we get signal from %s: %s", sender_name, signal_name);
    /* from org.ofono.MessageManager */
    if ((!strcmp(signal_name, "IncomingMessage")) ||
        (!strcmp(signal_name, "ImmediateMessage"))) {
        powerd_debug("Waking up the device - Incoming SMS");
        turn_display_on(TRUE, UNITY_SCREEN_REASON_NORMAL);
    }
}

void on_ofono_ussd_signal(GDBusProxy *proxy,
           gchar      *sender_name,
           gchar      *signal_name,
           GVariant   *parameters,
           gpointer    user_data)
{
    powerd_debug("we get signal from %s: %s", sender_name, signal_name);
    /* from org.ofono.MessageManager */
    if ((!strcmp(signal_name, "NotificationReceived")) ||
        (!strcmp(signal_name, "RequestReceived"))) {
        powerd_debug("Waking up the device - Incoming USSD");
        turn_display_on(TRUE, UNITY_SCREEN_REASON_NORMAL);
    }
}

/* Must be first to run on main loop */
static gboolean main_init(gpointer unused)
{
    powerd_mainloop_thread = g_thread_self();
    return FALSE;
}

/* Returns TRUE if current thread of execution is the default main loop */
int powerd_is_mainloop(void)
{
    return (g_thread_self() == powerd_mainloop_thread);
}

/* If running in emulator, make sure we keep the display on and
 * block suspend (can easily corrupt the data, not properly supported) */
static gboolean check_emulator(gpointer unused)
{
    char propbuf[PROP_VALUE_MAX];
    static uuid_t emulator_request_cookie;

    if (property_get("ro.kernel.qemu", propbuf, NULL) > 0) {
        powerd_info("Running in emulator, forcing display, blocking suspend.");
        keep_display_on();
        /* Block suspend */
        request_sys_state_internal("emulator-request",
                                  POWERD_SYS_STATE_ACTIVE,
                                  &emulator_request_cookie, NULL);
    }

    return FALSE;
}

void powerd_shutdown(void)
{
    static char poweroff_cmd[] = "poweroff";
    char *argv[] = {poweroff_cmd, NULL};
    g_spawn_async(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL);
}

void powerd_exit(int exit_code)
{
    g_exit_code = exit_code;
    g_main_loop_quit(main_loop);
}

int main(int argc, char** argv)
{
    powerd_log_init();

    powerd_info("Initializing powerd.");

    name_id = g_bus_own_name(G_BUS_TYPE_SYSTEM, "com.canonical.powerd",
        G_BUS_NAME_OWNER_FLAGS_REPLACE, powerd_bus_acquired_cb,
        powerd_name_acquired_cb, powerd_name_lost_cb, NULL, NULL);
    powerd_debug("owner id: %u", name_id);

    /* Listen to modem creation */
    g_dbus_proxy_new_for_bus(G_BUS_TYPE_SYSTEM,
        G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
        NULL,
        "org.ofono",
        "/",
        "org.ofono.Manager",
        NULL,
        (GAsyncReadyCallback) ofono_manager_proxy_connect_cb,
        NULL);

    /* Init this first, data is used by other inits */
    device_config_init();

    libsuspend_init(0);
    powerd_stats_init();
    powerd_client_init();
    power_request_init();
    dbus_name_watch_init();
    powerd_backlight_init();
    powerd_autobrightness_init();
    powerd_sensors_init();
    powerd_ps_init();

    main_loop = g_main_loop_new (NULL, FALSE);
    signal(SIGTERM, sigterm_quit);

    int err = hw_get_module(POWER_HARDWARE_MODULE_ID,
        (hw_module_t const**)&_power_module);

    if (!err)
        _power_module->init(_power_module);

    /* Config use should be done during init, okay to free now */
    device_config_deinit();

    /* This needs to be the first thing to run on the main loop */
    g_idle_add_full(G_PRIORITY_HIGH, main_init, NULL, NULL);
    g_idle_add(check_emulator, NULL);

    g_main_loop_run(main_loop);
    g_bus_unown_name(name_id);
    g_main_loop_unref(main_loop);

    powerd_ps_deinit();
    dbus_name_watch_deinit();
    powerd_autobrightness_deinit();
    powerd_backlight_deinit();
    power_request_deinit();
    powerd_client_deinit();
    powerd_stats_deinit();
    g_slist_free_full(g_modems, g_free);
    g_slist_free_full(g_calls, free_call_data);

    return g_exit_code;
}
